"""
Please note that all the functions from ``gempipe.curate.*`` are internally called also from ``gempipe``. 
Therefore, for example, to use the `perform_gapfilling()` function from ``gempipe.curate.gaps``, it's sufficient to write::

    import gempipe
    
    gempipe.perform_gapfilling(model, universe)
"""
